/* ifaddrs-compat.c -- implementation for getting network interface addresses
 Copyright (C) 2002 Free Software Foundation, Inc.
 This file is part of the GNU C Library.

 The GNU C Library is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.

 The GNU C Library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public
 License along with the GNU C Library; if not, write to the Free
 Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
 02111-1307 USA.  */

#include "system.h"

#if defined(HAVE_IFADDRS_H)
#include "device.h"
#include "ifaddrs.h"
#else
#include "ifaddrs-compat.h"
#include "net.h"
#include "xalloc.h"
#include <sys/types.h>
#include "logger.h"
#include <errno.h>
#include "utils.h"

# if !defined(IF_NAMESIZE)
#  if defined(IFNAMSIZ)
#   define IF_NAMESIZE IFNAMSIZ
#  else
#   define IFNAMSIZ 256
#  endif
#endif

#if defined(HAVE_MINGW)
#include <iphlpapi.h>
const ULONG flags = GAA_FLAG_INCLUDE_PREFIX | GAA_FLAG_SKIP_MULTICAST | GAA_FLAG_SKIP_ANYCAST;

struct ifreq {
	union {
		char ifrn_name[IFNAMSIZ]; /* if name, e.g. "en0" */
	}ifr_ifrn;

	union {
		struct sockaddr ifru_addr;
		struct sockaddr ifru_dstaddr;
		struct sockaddr ifru_broadaddr;
		struct sockaddr ifru_netmask;
		unsigned int ifru_flags;
		void * ifru_data;
	}ifr_ifru;
};

#define ifr_name        ifr_ifrn.ifrn_name      /* interface name         */
#define ifr_hwaddr      ifr_ifru.ifru_hwaddr    /* MAC address            */
#define ifr_addr        ifr_ifru.ifru_addr      /* address                */
/* other end of p-p lnk   */
#define ifr_broadaddr   ifr_ifru.ifru_broadaddr /* broadcast address      */
#define ifr_dstaddr     ifr_ifru.ifru_dstaddr
#define ifr_netmask     ifr_ifru.ifru_netmask   /* interface net mask     */
#define ifr_flags       ifr_ifru.ifru_flags     /* flags                  */
#define ifr_index		ifr_ifru.ifru_index		/* index of the interface */

#else
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <net/if.h>
#include <netinet/in.h>
#endif

#if defined(HAVE_SYS_SOCKIO_H)
#include <sys/sockio.h>
#elif defined(HAVE_LINUX_SOCKIOS_H)
#include <linux/sockios.h>
#endif

#define MAX_INTERFACES	128
#define STEPSZ			  8

static void add_interface(struct ifaddrs **ifa, struct ifreq *ifNameAddress, struct ifreq *ifFlags, int ifIndex, struct sockaddr *sa) {
	struct ifaddrs *ifa_new = NULL;
	size_t sasize;
	int family = ifNameAddress->ifr_addr.sa_family;

	if (family != AF_INET && family != AF_INET6)
		return;

	sasize = SALEN((*sa));

	if ((ifa_new = malloc(sizeof(*ifa_new))) == NULL)
		goto fail;
	memset(ifa_new, 0, sizeof(*ifa_new));

	if ((ifa_new->ifa_name = strdup(ifNameAddress->ifr_name)) == NULL)
		goto fail;
	if ((ifa_new->ifa_addr = malloc(sasize)) == NULL)
		goto fail;
	if (memcpy(ifa_new->ifa_addr, sa, sasize) == NULL)
		goto fail;
	ifa_new->ifa_flags = ifFlags->ifr_flags;
	ifa_new->ifa_index = ifIndex;

	ifa_new->ifa_next = *ifa;
	*ifa = ifa_new;

	return;
	fail: if (ifa_new->ifa_name != NULL)
		free(ifa_new->ifa_name);
	if (ifa_new->ifa_addr != NULL)
		free(ifa_new->ifa_addr);
	if (ifa_new != NULL)
		free(ifa_new);
}

void freeifaddrs(struct ifaddrs *ifap) {
	struct ifaddrs *tmpifa = ifap;

	while (ifap != NULL) {
		free(ifap->ifa_name);
		free(ifap->ifa_addr);

		tmpifa = ifap->ifa_next;
		free(ifap);
		ifap = tmpifa;
	}
	return;
}

#if defined(HAVE_MINGW)

int getifaddrs(struct ifaddrs **ifa) {
	/* Declare and initialize variables */
	DWORD dwSize = 0;
	DWORD dwRetVal = 0;
	char *afname = NULL, *address = NULL;

	unsigned int i = 0;

	LPVOID lpMsgBuf = NULL;

	PIP_ADAPTER_ADDRESSES pAddresses = NULL;
	ULONG outBufLen = 0;

	PIP_ADAPTER_ADDRESSES pCurrAddresses = NULL;
	PIP_ADAPTER_UNICAST_ADDRESS pUnicast = NULL;
	PIP_ADAPTER_ANYCAST_ADDRESS pAnycast = NULL;
	PIP_ADAPTER_MULTICAST_ADDRESS pMulticast = NULL;
	IP_ADAPTER_DNS_SERVER_ADDRESS *pDnServer = NULL;
	IP_ADAPTER_PREFIX *pPrefix = NULL;

	/* Determine required buffer size */
	do {
		free(pAddresses);
		pAddresses = (IP_ADAPTER_ADDRESSES *) xmalloc(outBufLen);
		dwRetVal = GetAdaptersAddresses(AF_UNSPEC, flags, NULL, pAddresses, &outBufLen);
		if (dwRetVal != ERROR_SUCCESS && dwRetVal != ERROR_BUFFER_OVERFLOW) {
			logger(LOG_ERR, "Call to %s failed: %d %s", "GetAdaptersAddresses", dwRetVal, winerror(dwRetVal));
		}
	} while (dwRetVal != ERROR_SUCCESS && dwRetVal != ERROR_NO_DATA);

	if (dwRetVal == ERROR_NO_DATA) {
		/* No data found */
		logger(LOG_ERR, "Call to %s failed: %d %s", "GetAdaptersAddresses", dwRetVal, winerror(dwRetVal));
		free(pAddresses);
		return -1;
	}
	pCurrAddresses = pAddresses;
	while (pCurrAddresses) {
		for (pUnicast = pCurrAddresses->FirstUnicastAddress; pUnicast != NULL;) {
			if (pUnicast->Flags & IP_ADAPTER_ADDRESS_DNS_ELIGIBLE == 0) {
				/* address should not be used to announce */
				pUnicast = pUnicast->Next;
				continue;
			}
			sockaddr_t sa;
			sa.sa.sa_family = pUnicast->Address.lpSockaddr->sa_family;
			if (sa.sa.sa_family == AF_INET) {
				sa.in = *((struct sockaddr_in *) pUnicast->Address.lpSockaddr);
			}
			if (sa.sa.sa_family == AF_INET6) {
				sa.in6 = *((struct sockaddr_in6 *) pUnicast->Address.lpSockaddr);
			}
			struct ifreq ifNameAddress, ifFlags;
			int ifIndex;
			strncpy((char*) &ifNameAddress.ifr_name, pCurrAddresses->AdapterName, IFNAMSIZ);
			ifNameAddress.ifr_addr.sa_family =
					pUnicast->Address.lpSockaddr->sa_family;
			ifFlags.ifr_flags = 0;
			ifFlags.ifr_flags |= ((pCurrAddresses->OperStatus & IfOperStatusUp) ? IFF_UP : 0);
			ifFlags.ifr_flags |= ((pCurrAddresses->IfType & IF_TYPE_SOFTWARE_LOOPBACK) ? IFF_LOOPBACK : 0);
			ifFlags.ifr_flags |= ((pCurrAddresses->Flags & IP_ADAPTER_NO_MULTICAST) ? 0 : IFF_MULTICAST);

			ifIndex = pCurrAddresses->IfIndex;
			add_interface(ifa, &ifNameAddress, &ifFlags, ifIndex, (struct sockaddr*) &sa);
			pUnicast = pUnicast->Next;
		}
		pCurrAddresses = pCurrAddresses->Next;
	}

	free(pAddresses);
	return 0;
}

#elif defined(HAVE_SYS_SOCKIO_H) || defined(HAVE_LINUX_SOCKIOS_H)

int getifaddrs(struct ifaddrs **ifa) {
	int fd, i, oldlen = 0;
	size_t buflen = 0;
	char *oldbuf = NULL, *buf = NULL;
	struct ifconf ifc;
	struct ifreq *ifr;
	struct sockaddr *sa;

	if ((fd = socket(AF_INET, SOCK_DGRAM, 0)) < 0) {
		logger(LOG_ERR, "getifaddrs: Call to &s failed", "socket");
		return -1;
	}

	*ifa = NULL;

	for (i = STEPSZ; i < MAX_INTERFACES; i += STEPSZ) {
		buflen = i * sizeof(struct ifreq);
		oldbuf = buf;
		if ((buf = xrealloc(buf, buflen)) == NULL) {
			logger(LOG_ERR, "getifaddrs: Call to &s failed", "xrealloc");
			goto fail;
		}
		ifc.ifc_buf = buf;
		ifc.ifc_len = buflen;

		if (ioctl(fd, SIOCGIFCONF, &ifc) < 0) {
			logger(LOG_ERR, "getifaddrs: Call to &s failed", "SIOCGIFCONF");
			goto fail;
		}
		if (oldlen == ifc.ifc_len)
		break;
		oldlen = ifc.ifc_len;
	}

	for (oldbuf = buf; buf < oldbuf + ifc.ifc_len;) {
		ifr = (struct ifreq *)buf;
		sa = (struct sockaddr *)&ifr->ifr_addr;
		buf += IF_NAMESIZE + SALEN((*sa));

		struct ifreq ifFlags = *ifr;
		struct ifreq ifIndex = *ifr;
		if (ioctl(fd, SIOCGIFFLAGS, &ifFlags) < 0) {
			logger(LOG_ERR, "getifaddrs: Call to &s failed", "SIOCGIFFLAGS");
			goto fail;
		}
		if (ioctl(fd, SIOCGIFINDEX, &ifIndex) < 0) {
			logger(LOG_ERR, "getifaddrs: Call to &s failed", "SIOCGIFINDEX");
			goto fail;
		}

		add_interface(ifa, ifr, &ifFlags, ifIndex.ifr_ifindex, sa);
	}

	free(buf);
	close(fd);
	return 0;

fail:
	free(oldbuf);
	if (*ifa != NULL) {
		freeifaddrs(*ifa);
		*ifa = NULL;
	}
	close(fd);
	return -1;
}
# endif

#endif /* HAVE_GETIFADDRS */
