/*
 * multicast.c
 *
 *  Created on: 05.07.2010
 *      Author: schalli
 */

#include "system.h"

#include <openssl/hmac.h>

#include "ifaddrs-compat.h"

#include "device.h"
#include "multicast.h"
#include "netutl.h"

#ifdef HAVE_MINGW
#include <ws2tcpip.h>
#endif

#include <lib/xalloc.h>

const int multicastTTL = 1;

const int multicastPort = 16655;

struct sockaddr_in multicastAddr4;

struct sockaddr_in6 multicastAddr6;

listen_socket_t multicast_socket[MAXSOCKETS];
int multicast_sockets = 0;

char mc_challenge[MC_CHALLENGE_SIZE] = { 0 };
struct timeval mc_timestamp;

/** Check MAC of incoming multicast packet */
int check_signature(const node_t *n, mcpacket_t *packet) {
	if (memcmp(mc_challenge, (char *) &packet->challenge, sizeof(MC_CHALLENGE_SIZE))) {
		ifdebug(PROTOCOL)
			logger(LOG_DEBUG, "Got bogus multicast response from %s", n->name);
		return 0;
	}
	if (n->indigest && n->inmaclength) {
		unsigned char hmac[EVP_MAX_MD_SIZE];
		HMAC(n->indigest, n->inkey, n->inkeylength, (unsigned char *) packet, (char *) &packet->response.signature - (char *) packet, (unsigned char *) &hmac,
				NULL);
		if (memcmp(hmac, (char *) &packet->response.signature, n->inmaclength)) {
			ifdebug(PROTOCOL)
				logger(LOG_DEBUG, "Got unauthenticated multicast response from %s", n->name);
			return 0;
		}
		// challenge and signature ok
		return 1;
	} else {
		ifdebug(PROTOCOL)
			logger(LOG_WARNING, "Unable to check signature in multicast response from %s", n->name);
		return 0;
	}
}

/* Add the message authentication code */
void set_signature(const node_t *n, mcpacket_t* packet) {
	if (n->outdigest && n->outmaclength) {
		HMAC(n->outdigest, n->outkey, n->outkeylength, (unsigned char *) packet, (char *) &packet->response.signature - (char *) packet,
				(unsigned char *) &packet->response.signature, NULL);
	} else {
		ifdebug(PROTOCOL)
			logger(LOG_WARNING, "Unable to sign multicast response to %s", n->name);
	}
}

mcpacket_t *new_mcpacket(void) {
	return xmalloc_and_zero(sizeof(mcpacket_t));
}

void free_mcpacket(mcpacket_t *p) {
	free(p);
}

void init_multicast(void) {
	/** Initialize */
	const char *multicastIP4 = "224.0.0.1";
	multicastAddr4 = str2sockaddr(multicastIP4, 0x0).in;
	multicastAddr4.sin_port = htons(multicastPort); /* Multicast port */

	const char *multicastIP6 = "FF02:0:0:0:0:0:0:1";
	multicastAddr6 = str2sockaddr(multicastIP6, 0x0).in6;
	multicastAddr6.sin6_port = htons(multicastPort); /* Multicast port */
	/** Done initializing */
}

/** Bind to sockets */
void bind_multicast_sockets(void) {
	char *mode = NULL;
	if (get_config_string(lookup_config(config_tree, "LocalAnnounce"), &mode)) {
		if (strcasecmp(mode, "ReceiveOnly") && strcasecmp(mode, "AnnounceAndReceive")) {
			free(mode);
			return;
		}
	} else {
		return;
	}
	free(mode);

	int afamily = AF_UNSPEC;
	char *afname = NULL, *address = NULL;
	multicast_sockets = 0;

	if (get_config_string(lookup_config(config_tree, "AddressFamily"), &afname)) {
		if (!strcasecmp(afname, "IPv4"))
			afamily = AF_INET;
		else if (!strcasecmp(afname, "IPv6"))
			afamily = AF_INET6;
		else if (!strcasecmp(afname, "any"))
			afamily = AF_UNSPEC;
		else {
			logger(LOG_ERR, "Invalid address family!");
			free(afname);
			return;
		}
		free(afname);
	}

	if (get_config_string(lookup_config(config_tree, "BindToAddress"), &address)) {
		logger(LOG_ERR, "BindToAddress not yet supported");
		free(address);
		return;
	} else {
		// bind to all interfaces
		struct ifaddrs *ifaddr = NULL, *ifa;
		int family, s;
		char host[NI_MAXHOST];
		if (getifaddrs(&ifaddr) < 0) {
			logger(LOG_ERR, "bind_multicast_sockets: Call to %s failed", "getifaddrs");
			return;
		}

		for (ifa = ifaddr; ifa != NULL; ifa = ifa->ifa_next) {
			family = ifa->ifa_addr->sa_family;

			/* For an AF_INET* interface address, display the address */
			if (family != AF_INET && family != AF_INET6) {
				// neither ipv4 nor ipv6
				continue;
			}
			if (!(ifa->ifa_flags & IFF_UP)) {
				// interface is down
				continue;
			}

			if (ifa->ifa_flags & IFF_LOOPBACK) {
				// loopback
				continue;
			}

			if (ifa->ifa_flags & IFF_POINTTOPOINT) {
				// p2p
				continue;
			}
			if (!(ifa->ifa_flags & IFF_MULTICAST)) {
				// does not support multicast
				continue;
			}
			if (strcmp(iface, ifa->ifa_name) == 0) {
				// it's the tinc interface
				continue;
			}

			sockaddr_t sa;
			memcpy(&sa, ifa->ifa_addr, SALEN((*ifa->ifa_addr)));
			//char* address;
			//sockaddr2str(&sa, &address, NULL);
			//printf("Binding %s\n", address);
			//free(address);

			int ifIndex;
#ifdef HAVE_MINGW
			ifIndex = ifa->ifa_index;
#else
			ifIndex = if_nametoindex(ifa->ifa_name);
#endif
			if (ifIndex == 0) {
				logger(LOG_WARNING, "Failed to call if_nametoindex()");
				continue;
			}
			bind_multicast_socket(&sa, ifIndex);
		}

		freeifaddrs(ifaddr);
	}
}

void bind_multicast_socket(sockaddr_t *sa, int ifIndex) {
	int fd;
	fd = socket(sa->sa.sa_family, SOCK_DGRAM, 0);
	if (fd < 0) {
		ifdebug(PROTOCOL)
			logger(LOG_WARNING, "Creating ipv4 multicast listener socket failed: %s", sockstrerror(sockerrno));
		return;
	}

	/* Double bind to address */
	int option = 1;
	if (setsockopt(fd, SOL_SOCKET, SO_REUSEADDR, (void *) &option, sizeof(option)) < 0) {
		ifdebug(PROTOCOL)
			logger(LOG_WARNING, "Could not set socket options: %s", sockstrerror(sockerrno));
		closesocket(fd);
		return;
	}

	if (sa->sa.sa_family == AF_INET) {
		sa->in.sin_port = htons(multicastPort); /* Multicast port */

		/* Bind to the multicast port */
		if (bind(fd, (struct sockaddr *) &sa->in, sizeof(sa->in)) < 0) {
			ifdebug(PROTOCOL)
				logger(LOG_WARNING, "Could not bind to ipv4 multicast socket: %d %s", sockerrno, sockstrerror(sockerrno));
			closesocket(fd);
			return;
		}

		/* Join the multicast address */
		struct ip_mreq multicastRequest;
		multicastRequest.imr_multiaddr = multicastAddr4.sin_addr;
#ifdef HAVE_MINGW
		multicastRequest.imr_interface.s_addr = sa->in.sin_addr.S_un.S_addr;
#else
		multicastRequest.imr_interface = sa->in.sin_addr;
#endif
		if (setsockopt(fd, IPPROTO_IP, IP_ADD_MEMBERSHIP, (char *) &multicastRequest, sizeof(multicastRequest)) < 0) {
			ifdebug(PROTOCOL)
				logger(LOG_WARNING, "Could not join ipv4 multicast group: %d %s", sockerrno, sockstrerror(sockerrno));
			closesocket(fd);
			return;
		}
	} else if (sa->sa.sa_family == AF_INET6) {
		sa->in6.sin6_port = htons(multicastPort); /* Multicast port */

		/* Bind to the multicast port */
		if (bind(fd, (struct sockaddr *) &sa->in6, sizeof(sa->in6)) < 0) {
			ifdebug(PROTOCOL)
				logger(LOG_WARNING, "Could not bind to ipv6 multicast socket: %d %s", sockerrno, sockstrerror(sockerrno));
			closesocket(fd);
			return;
		}

		/* Join the multicast address */
		struct ipv6_mreq multicastRequest;
		multicastRequest.ipv6mr_multiaddr = multicastAddr6.sin6_addr;
		multicastRequest.ipv6mr_interface = ifIndex;
		if (setsockopt(fd, IPPROTO_IPV6, IPV6_ADD_MEMBERSHIP, (char *) &multicastRequest, sizeof(multicastRequest)) < 0) {
			ifdebug(PROTOCOL)
				logger(LOG_WARNING, "Could not join ipv6 multicast group: %d %s", sockerrno, sockstrerror(sockerrno));
			closesocket(fd);
			return;
		}
	} else {
		ifdebug(PROTOCOL)
			logger(LOG_WARNING, "Cannot announce endpoint: %s", "Unknown address family");
		closesocket(fd);
		return;
	}

	multicast_socket[multicast_sockets].sa = *sa;
	multicast_socket[multicast_sockets++].udp = fd;
}

void handle_multicast_data(int sock) {
	length_t len;
	sockaddr_t from;
	socklen_t fromlen = sizeof(from);
	mcpacket_t packet;

	len = recvfrom(sock, (char *) &packet, sizeof(mcpacket_t), 0, &from.sa, &fromlen);

	if (len != sizeof(mcpacket_t)) {
		logger(LOG_INFO, "Received bogus multicast data");
		return;
	}
	receive_multicast_packet(sock, &packet, &from);
}

void receive_multicast_packet(int fd, mcpacket_t *packet, sockaddr_t *from) {
	if (strncmp(packet->netname, netname, STRLEN) != 0) {
		/* wrong netname */
		return;
	}

	if (strncmp(packet->hostname, myself->name, STRLEN) == 0) {
		/* from myself --> ignore */
		return;
	}

	node_t *n = lookup_node(packet->hostname);
	if (!n) {
		/* node not found */
		logger(LOG_WARNING, "Node %s not found", packet->hostname);
		return;
	}

	//	printf("Receiving Multicast Data\n");
	if (packet->type == MCPKT_ANNOUNCE) {
		//		printf("Multicast Announce reveived from %s\n", packet->hostname);
		receive_multicast_announce(fd, packet, n, from);
	} else if (packet->type == MCPKT_RESPONSE) {
		//		printf("Multicast Response reveived from %s\n", packet->hostname);
		if (!check_signature(n, packet)) {
			/* Packet not properly signed */
			logger(LOG_WARNING, "Packet from %s not properly signed", n->name);
			return;
		}

		receive_multicast_response(n, from, packet);
	} else {
		//		ifdebug(PROTOCOL)
		logger(LOG_WARNING, "Unknown Multicast Packet received from %s", n->name);
	}
}

void close_multicast_sockets() {
	for (int i = 0; i < multicast_sockets; i++) {
		close_multicast_socket(i);
	}
	multicast_sockets = 0;
}

void close_multicast_socket(const int i) {
	close(multicast_socket[i].udp);
	multicast_socket[i].udp = -1;
	//	for (int i = socketIdx + 1; i < multicast_sockets; i++) {
	//		multicast_socket[i - 1] = multicast_socket[i];
	//	}
	//	multicast_sockets--;
}
